defmodule AliyunAPI.OSS do
  require Logger
  import AliyunAPI, only: [url_encode: 1]
  import SweetXml

  ##### Start 分片上传 ####################
  # https://help.aliyun.com/document_detail/55538.html

  @doc """
  使用Multipart Upload模式传输数据前，必须先调用该接口来通知OSS初始化一个Multipart Upload事件。
  https://help.aliyun.com/document_detail/31992.html

  ## Examples

      iex> AliyunAPI.OSS.initiate_multipart_upload(:default, "multipart.data")

  """
  def initiate_multipart_upload(bucket_name, object_key, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    opts =
      opts
      |> Keyword.put(:verb, "POST")
      |> Keyword.put(:content_type, "application/octet-stream")
      |> Keyword.put(
        :canonicalized_resource,
        Enum.join(["/#{bucket_name}", object_key <> "?uploads"], "/")
      )

    [bucket_url, object_key <> "?uploads"]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:post, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, headers: headers, body: body}} ->
        Logger.info("initiate_multipart_upload##{object_key}@#{bucket_name}", filter: :oss)

        data =
          body
          |> SweetXml.xpath(~x"//InitiateMultipartUploadResult",
            bucket: ~x"./Bucket/text()"so,
            key: ~x"./Key/text()"so,
            upload_id: ~x"./UploadId/text()"so
          )

        {:ok, 200, %{headers: headers, body: body, data: data}}

      {:ok, %HTTPoison.Response{status_code: status_code, headers: headers, body: body}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so,
            request_id: ~x"./RequestId/text()"so,
            host_id: ~x"./HostId/text()"so,
            bucket: ~x"./Bucket/text()"so,
            endpoint: ~x"./Endpoint/text()"so
          )

        {:error, status_code, %{body: body, headers: headers, error: error}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  AbortMultipartUpload接口可以根据用户提供的Upload ID中止其对应的Multipart Upload事件。
  当一个Multipart Upload事件被中止后，就不能再使用这个Upload ID做任何操作，已经上传的Part数据也会被删除。
  https://help.aliyun.com/document_detail/31996.html
  """
  def abort_multipart_upload(bucket_name, object_key, upload_id, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    opts =
      opts
      |> Keyword.put(:verb, "DELETE")
      |> Keyword.put(
        :canonicalized_resource,
        Enum.join(["/#{bucket_name}", object_key <> "?uploadId=#{upload_id}"], "/")
      )

    [bucket_url, object_key <> "?uploadId=#{upload_id}"]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:delete, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 204, headers: headers, body: body}} ->
        Logger.info("abort_multipart_upload##{object_key}@#{bucket_name}", filter: :oss)
        {:ok, 204, %{headers: headers, body: body, data: %{}}}

      {:ok, %HTTPoison.Response{status_code: status_code, headers: headers, body: body}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so,
            request_id: ~x"./RequestId/text()"so,
            host_id: ~x"./HostId/text()"so,
            bucket: ~x"./Bucket/text()"so,
            endpoint: ~x"./Endpoint/text()"so
          )

        {:error, status_code, %{body: body, headers: headers, error: error}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  初始化一个Multipart Upload之后，可以根据指定的Object名和Upload ID来分块（Part）上传数据。每一个上传的Part都有一个标识它的号码（part number，范围是1~10,000）。
  https://help.aliyun.com/document_detail/31993.html

  * bucket_name
  * object_key
  * opts
    - :`part_number`
    - :`upload_id`
    - :`part`
  """
  def upload_part(bucket_name, object_key, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    query_string =
      URI.encode_query(
        partNumber: opts[:part_number],
        uploadId: opts[:upload_id]
      )

    opts =
      opts
      |> Keyword.put(:verb, "PUT")
      |> Keyword.put(:content_type, "application/octet-stream")
      |> Keyword.put(
        :canonicalized_resource,
        Enum.join(["/#{bucket_name}", object_key <> "?#{query_string}"], "/")
      )

    [bucket_url, object_key <> "?#{query_string}"]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:put, body: opts[:part], headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, headers: headers, body: body}} ->
        Logger.info("upload_part##{object_key}@#{bucket_name}", filter: :oss)

        etag =
          AliyunAPI.get_header(headers, :etag)
          |> List.first()
          |> to_string()
          |> String.trim("\"")
          |> String.trim()

        {:ok, 200, %{headers: headers, body: body, data: %{etag: etag}}}

      {:ok, %HTTPoison.Response{status_code: status_code, headers: headers, body: body}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so,
            host_id: ~x"./HostId/text()"so,
            upload_id: ~x"./UploadId/text()"so
          )

        {:error, status_code, %{body: body, headers: headers, error: error}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  在将所有数据Part都上传完成后，必须调用Complete Multipart Upload API来完成整个文件的Multipart Upload。
  https://help.aliyun.com/document_detail/31995.html

  * bucket_name
  * object_key
  * upload_id
  * parts
  * opts
  """
  def complete_multipart_upload(bucket_name, object_key, upload_id, parts, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    parts_body =
      XmlBuilder.document([
        XmlBuilder.element(:CompleteMultipartUpload, %{}, [
          Enum.map(parts || [], fn part ->
            XmlBuilder.element(:Part, %{}, [
              XmlBuilder.element(:PartNumber, %{}, part[:number]),
              XmlBuilder.element(:ETag, %{}, part[:etag])
            ])
          end)
        ])
      ])
      |> XmlBuilder.generate()

    opts =
      opts
      |> Keyword.put(:verb, "POST")
      |> Keyword.put(:content_type, "application/octet-stream")
      |> Keyword.put(
        :canonicalized_resource,
        Enum.join(["/#{bucket_name}", object_key <> "?uploadId=#{upload_id}"], "/")
      )

    [bucket_url, object_key <> "?uploadId=#{upload_id}"]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:post, body: parts_body, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, headers: headers, body: body}} ->
        Logger.info("complete_multipart_upload##{object_key}@#{bucket_name}", filter: :oss)

        data =
          body
          |> SweetXml.xpath(~x"//CompleteMultipartUploadResult",
            bucket: ~x"./Bucket/text()"so,
            location: ~x"./Location/text()"so,
            key: ~x"./Key/text()"so,
            etag: ~x"./ETag/text()"so
          )

        {:ok, 200, %{headers: headers, body: body, data: data}}

      {:ok, %HTTPoison.Response{status_code: status_code, headers: headers, body: body}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so,
            host_id: ~x"./HostId/text()"so,
            upload_id: ~x"./UploadId/text()"so
          )

        {:error, status_code, %{body: body, headers: headers, error: error}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  在将所有数据Part都上传完成后，必须调用Complete Multipart Upload API来完成整个文件的Multipart Upload。
  https://help.aliyun.com/document_detail/31995.html

  * bucket_name
  * opts
    - :`delimiter`
    - :`max_uploads`
    - :`key_marker`
    - :`prefix`
    - :`upload_id_marker`
    - :`encoding_type`

  ## Examples

      iex> AliyunAPI.OSS.list_multipart_uploads(:default)

  """
  def list_multipart_uploads(bucket_name, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    query_string =
      %{
        delimiter: opts[:delimiter],
        maxUploads: opts[:max_uploads],
        keyMarker: opts[:key_marker],
        prefix: opts[:prefix],
        uploadIdMarker: opts[:upload_id_marker],
        encodingType: opts[:encoding_type]
      }
      |> _encode_arguments()

    opts =
      opts
      |> Keyword.put(:verb, "GET")
      |> Keyword.put(
        :canonicalized_resource,
        "/#{bucket_name}/?uploads"
      )

    [bucket_url, "?uploads#{if String.length(query_string) > 0, do: "&#{query_string}"}"]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:get, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, headers: headers, body: body}} ->
        Logger.info("list_multipart_uploads@#{bucket_name}", filter: :oss)

        data =
          body
          |> SweetXml.xpath(~x"//ListMultipartUploadsResult",
            bucket: ~x"./Bucket/text()"so,
            key_marker: ~x"./KeyMarker/text()"so,
            upload_id_marker: ~x"./UploadIdMarker/text()"so,
            next_key_marker: ~x"./NextKeyMarker/text()"so,
            next_upload_id_marker: ~x"./NextUploadIdMarker/text()"so,
            key_marker: ~x"./KeyMarker/text()"so,
            delimiter: ~x"./Delimiter/text()"so,
            prefix: ~x"./Prefix/text()"so,
            max_uploads: ~x"./MaxUploads/text()"io,
            is_truncated: ~x"./IsTruncated/text()"so,
            uploads: [
              ~x".//Upload"lo,
              key: ~x"./Key/text()"so,
              upload_id: ~x"./UploadId/text()"so,
              initiated: ~x"./Initiated/text()"so
            ]
          )

        {:ok, 200, %{headers: headers, body: body, data: data}}

      {:ok, %HTTPoison.Response{status_code: status_code, headers: headers, body: body}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so,
            host_id: ~x"./HostId/text()"so
          )

        {:error, status_code, %{body: body, headers: headers, error: error}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  ListParts接口用于列举指定Upload ID所属的所有已经上传成功Part。
  https://help.aliyun.com/document_detail/31998.html

  * bucket_name
  * object_key
  * upload_id
  * opts
    - :`max_parts`
    - :`part_number_marker`
    - :`encoding_type`

  ## Examples

      iex> AliyunAPI.OSS.list_parts(:default, "multipart.data", "3D76F1DC072A4BB0845A808EBE61C7AC")

  """
  def list_parts(bucket_name, object_key, upload_id, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    query_string =
      %{
        uploadId: upload_id,
        maxParts: opts[:max_parts],
        partNumberMarker: opts[:part_number_marker],
        encodingType: opts[:encoding_type]
      }
      |> _encode_arguments()

    opts =
      opts
      |> Keyword.put(:verb, "GET")
      |> Keyword.put(
        :canonicalized_resource,
        Enum.join(["/#{bucket_name}", object_key <> "?uploadId=#{upload_id}"], "/")
      )

    [bucket_url, object_key <> "?#{query_string}"]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:get, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, headers: headers, body: body}} ->
        Logger.info("list_parts##{object_key}@#{bucket_name}", filter: :oss)

        data =
          body
          |> SweetXml.xpath(~x"//ListPartsResult",
            bucket: ~x"./Bucket/text()"so,
            key: ~x"./Key/text()"so,
            upload_id: ~x"./UploadId/text()"so,
            next_part_number_marker: ~x"./NextPartNumberMarker/text()"io,
            max_parts: ~x"./MaxParts/text()"io,
            is_truncated: ~x"./IsTruncated/text()"so,
            parts: [
              ~x".//Part"lo,
              number: ~x"./PartNumber/text()"so,
              last_modified: ~x"./LastModified/text()"so,
              etag: ~x"./ETag/text()"so,
              size: ~x"./Size/text()"io
            ]
          )

        {:ok, 200, %{headers: headers, body: body, data: data}}

      {:ok, %HTTPoison.Response{status_code: status_code, headers: headers, body: body}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so,
            request_id: ~x"./RequestId/text()"so,
            host_id: ~x"./HostId/text()"so,
            bucket: ~x"./Bucket/text()"so,
            endpoint: ~x"./Endpoint/text()"so
          )

        {:error, status_code, %{body: body, headers: headers, error: error}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  ##### END 分片上传 ####################

  @doc """
  获取头信息

  * bucket_name
  * object_key
  * opts
    - :`content_type`
    - :`headers`
    - :`process`
    - :`oss_endpoint`
    - :`access_key_secret`
    - :`access_key_id`
    - :`security_token`
  """
  def head_object(bucket_name, object_key, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    opts =
      opts
      |> Keyword.put(:verb, "HEAD")
      |> Keyword.put(:canonicalized_resource, Enum.join(["/#{bucket_name}", object_key], "/"))

    [bucket_url, object_key]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:head, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, headers: headers}} ->
        Logger.info("head_object##{object_key}@#{bucket_name}", filter: :oss)
        {:ok, 200, %{headers: headers}}

      {:ok, %HTTPoison.Response{status_code: status_code, headers: headers, body: body}} ->
        {:error, status_code, %{body: body, headers: headers}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  获取头信息

  * bucket_name
  * object_key
  * opts
    - :`content_type`
    - :`headers`
    - :`process`
    - :`oss_endpoint`
    - :`access_key_secret`
    - :`access_key_id`
    - :`security_token`
  """
  def get_object_meta(bucket_name, object_key, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    opts =
      opts
      |> Keyword.put(:verb, "HEAD")
      |> Keyword.put(
        :canonicalized_resource,
        Enum.join(["/#{bucket_name}", object_key <> "?objectMeta"], "/")
      )

    [bucket_url, object_key <> "?objectMeta"]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:head, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, headers: headers}} ->
        Logger.info("get_object_meta##{object_key}@#{bucket_name}", filter: :oss)
        {:ok, 200, %{headers: headers}}

      {:ok, %HTTPoison.Response{status_code: status_code, headers: headers, body: body}} ->
        {:error, status_code, %{body: body, headers: headers}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  提交数据

  https://help.aliyun.com/document_detail/31978.html

  ## Params

  * bucket_name
  * object_key
  * opts
    - :`file`           【必须】
    - :`content_type`   【必须】
    - :`content_length` 【必须】
    - :`content_md5`    【必须】
    - :`headers`
    - :`oss_endpoint`
    - :`access_key_secret`
    - :`access_key_id`
    - :`security_token`
  """
  def put_object(bucket_name, object_key, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    opts =
      opts
      |> Keyword.put(:verb, "PUT")
      |> Keyword.put(:canonicalized_resource, Enum.join(["/#{bucket_name}", object_key], "/"))
      |> Keyword.put(:content_length, String.length(opts[:file]))

    [bucket_url, object_key]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:put, body: opts[:file], headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, body: body, headers: headers}} ->
        Logger.info("put_object##{object_key}@#{bucket_name}", filter: :oss)
        {:ok, 200, %{body: body, headers: headers}}

      {:ok, %HTTPoison.Response{status_code: status_code, body: body, headers: headers}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so
          )

        {:error, status_code, %{error: error, body: body, headers: headers}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  设置object的访问权限

  https://help.aliyun.com/document_detail/31986.html

  * bucket_name
  * object_key
  * opts
    - :`acl`
      可选值
      - private
      - public-read
      - public-read-write
      - default
  """
  def put_object_acl(bucket_name, object_key, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)
    acl = opts[:acl] || "default"

    opts =
      opts
      |> Keyword.put(:verb, "PUT")
      |> Keyword.put(:content_type, "application/octet-stream")
      |> Keyword.put(
        :canonicalized_resource,
        Enum.join(["/#{bucket_name}", object_key <> "?acl"], "/")
      )

    headers = opts |> Keyword.get(:headers, [])
    headers = headers ++ [{"x-oss-object-acl", acl}]
    opts = opts |> Keyword.put(:headers, headers)

    [bucket_url, object_key <> "?acl"]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:put, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, body: body, headers: headers}} ->
        Logger.info("put_object_acl##{object_key}@#{bucket_name}", filter: :oss)
        {:ok, 200, %{body: body, headers: headers}}

      {:ok, %HTTPoison.Response{status_code: status_code, body: body, headers: headers}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so
          )

        {:error, status_code, %{error: error, body: body, headers: headers}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  获取object的访问权限

  https://help.aliyun.com/document_detail/31986.html

  * bucket_name
  * object_key
  """
  def get_object_acl(bucket_name, object_key, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    opts =
      opts
      |> Keyword.put(:verb, "GET")
      |> Keyword.put(:content_type, "application/octet-stream")
      |> Keyword.put(
        :canonicalized_resource,
        Enum.join(["/#{bucket_name}", object_key <> "?acl"], "/")
      )

    [bucket_url, object_key <> "?acl"]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:get, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, body: body, headers: headers}} ->
        Logger.info("get_object_acl##{object_key}@#{bucket_name}", filter: :oss)

        data =
          body
          |> SweetXml.xpath(~x"//AccessControlPolicy",
            id: ~x"./Owner/ID/text()"io,
            display_name: ~x"./Owner/DisplayName/text()"so,
            grant: ~x"./AccessControlList/Grant/text()"so
          )

        {:ok, 200, %{body: body, data: data, headers: headers}}

      {:ok, %HTTPoison.Response{status_code: status_code, body: body, headers: headers}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so
          )

        {:error, status_code, %{error: error, body: body, headers: headers}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  删除数据

  * bucket_name
  * object_key
  * opts
    - :`headers`
    - :`oss_endpoint`
    - :`access_key_secret`
    - :`access_key_id`
    - :`security_token`
  """
  def delete_object(bucket_name, object_key, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    opts =
      opts
      |> Keyword.put(:verb, "DELETE")
      |> Keyword.put(:canonicalized_resource, Enum.join(["/#{bucket_name}", object_key], "/"))

    [bucket_url, object_key]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:delete, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 204, body: body, headers: headers}} ->
        Logger.info("delete_object##{object_key}@#{bucket_name}", filter: :oss)
        {:ok, 204, %{body: body, headers: headers}}

      {:ok, %HTTPoison.Response{status_code: status_code, body: body, headers: headers}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so
          )

        {:error, status_code, %{error: error, body: body, headers: headers}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  读取数据

  * bucket_name
  * object_key
  * opts
    - :`content_type`
    - :`headers`
    - :`process`
    - :`oss_endpoint`
    - :`access_key_secret`
    - :`access_key_id`
    - :`security_token`
  """
  def get_object(bucket_name, object_key, opts \\ []) do
    bucket_name = _get_bucket_name(opts[:app], bucket_name)
    bucket_url = bucket_name |> _get_bucket_url(opts)

    opts =
      opts
      |> Keyword.put(:verb, "GET")
      |> Keyword.put(:canonicalized_resource, Enum.join(["/#{bucket_name}", object_key], "/"))

    [bucket_url, object_key]
    |> Enum.join("/")
    |> _decorate_url(opts)
    |> AliyunAPI.http_request(:get, headers: compose_headers(opts))
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, body: body, headers: headers}} ->
        Logger.info("get_object##{object_key}@#{bucket_name}", filter: :oss)
        {:ok, 200, %{body: body, headers: headers}}

      {:ok, %HTTPoison.Response{status_code: status_code, body: body, headers: headers}} ->
        error =
          body
          |> SweetXml.xpath(~x"//Error",
            code: ~x"./Code/text()"so,
            message: ~x"./Message/text()"so
          )

        {:error, status_code, %{error: error, body: body, headers: headers}}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  # """
  # 生成签名 https://help.aliyun.com/document_detail/31951.html

  # * opts
  #   - :`headers`
  #   - :`verb`
  #   - :`content_md5`
  #   - :`content_type`
  #   - :`gmt_date`
  #   - :`access_key_secret`
  #   - :`security_token`
  # """
  defp _get_signature(opts) do
    headers = opts |> Keyword.get(:headers, [])
    access_key_secret = opts[:access_key_secret] || AliyunAPI.Env.oss_access_key_secret(opts[:app])

    oss_headers =
      headers
      |> Enum.reduce([], fn {k, v}, oss_headers ->
        if k |> String.downcase() |> String.starts_with?(["x-oss-"]) do
          oss_headers ++ ["#{String.downcase(k)}:#{v}"]
        else
          oss_headers
        end
      end)

    oss_headers = if Enum.any?(oss_headers), do: oss_headers, else: [""]

    canonicalized_oss_headers = oss_headers |> Enum.sort_by(fn kv -> kv end)
    canonicalized_resource = opts |> Keyword.get(:canonicalized_resource)

    signature_value =
      [
        opts[:verb],
        opts[:content_md5],
        opts[:content_type],
        opts[:gmt_date]
      ]
      |> AliyunAPI.if_call(Enum.any?(canonicalized_oss_headers), fn data ->
        data ++ canonicalized_oss_headers
      end)
      |> AliyunAPI.if_call(true, fn data ->
        data ++ ["#{canonicalized_resource}"]
      end)
      |> Enum.join("\n")

    :sha
    |> :crypto.hmac(access_key_secret, signature_value)
    |> :base64.encode()
  end

  defp compose_headers(opts) do
    gmt_date = Timex.now() |> Timex.format!("%a, %d %b %Y %H:%M:%S GMT", :strftime)
    canonicalized_resource = opts |> Keyword.get(:canonicalized_resource)
    access_key_id = opts[:access_key_id] || AliyunAPI.Env.oss_access_key_id(opts[:app])
    access_key_secret = opts[:access_key_secret] || AliyunAPI.Env.oss_access_key_secret(opts[:app])
    headers = opts |> Keyword.get(:headers, [])

    headers =
      (headers ++
         [
           {"Date", gmt_date},
           {"X-OSS-Meta-Version", "1.0"}
         ])
      |> AliyunAPI.if_call(!!opts[:security_token], fn headers ->
        headers ++ [{"x-oss-security-token", opts[:security_token]}]
      end)

    signature =
      _get_signature(
        canonicalized_resource: canonicalized_resource,
        verb: opts[:verb],
        content_md5: opts[:content_md5],
        content_type: opts[:content_type],
        access_key_secret: access_key_secret,
        gmt_date: gmt_date,
        headers: headers
      )

    (headers ++
       [
         {"Authorization", "OSS #{access_key_id}:#{signature}"}
       ])
    |> AliyunAPI.if_call(opts[:content_type], fn headers ->
      headers ++ [{"Content-Type", opts[:content_type]}]
    end)
    |> AliyunAPI.if_call(opts[:content_length], fn headers ->
      headers ++ [{"Content-Length", opts[:content_length]}]
    end)
    |> AliyunAPI.if_call(opts[:content_md5], fn headers ->
      headers ++ [{"Content-MD5", opts[:content_md5]}]
    end)
  end

  defp _get_bucket_name(app, bucket_name) when is_atom(bucket_name) do
    AliyunAPI.Env.oss_buckets(app) |> Map.get(bucket_name)
  end

  defp _get_bucket_name(_app, bucket_name), do: bucket_name

  defp _get_bucket_url(bucket_name, opts) do
    uri = "#{opts[:oss_endpoint] || AliyunAPI.Env.oss_endpoint(opts[:app])}" |> URI.parse()
    scheme = uri.scheme || "http"

    if uri.host && uri.scheme do
      %URI{uri | scheme: scheme, host: "#{bucket_name}.#{uri.host}"}
    else
      %URI{uri | scheme: scheme, host: "#{bucket_name}.#{uri.path}", path: nil}
    end
    |> URI.to_string()
  end

  defp _decorate_url(url, opts) do
    params =
      %{}
      |> AliyunAPI.if_call(!!opts[:process], fn params ->
        params |> Map.put("x-oss-process", opts[:process])
      end)

    url
    |> AliyunAPI.if_call(Enum.any?(params), fn url ->
      url <> "?" <> URI.encode_query(params)
    end)
  end

  defp _encode_arguments(arguments) do
    arguments
    |> Enum.filter(fn {_key, val} ->
      val |> to_string() |> String.length() > 0
    end)
    |> Enum.map(fn {key, val} -> {url_encode(key), url_encode(val)} end)
    |> Enum.map(fn {key, val} -> "#{key}=#{val}" end)
    |> Enum.join("&")
  end
end
