defmodule AliyunAPI.STS do
  import AliyunAPI, only: [url_encode: 1]
  require Logger

  @doc """
  通过该接口，获取一个扮演该角色的临时身份。

  https://help.aliyun.com/document_detail/28763.html

  ## opts

  * :`role_arn` String [必填]
  * :`role_session_name` String [必填]
  * :`policy` 策略
  * :`duration_seconds` 指定的过期时间，单位为秒。过期时间范围：900 ~ 3600，默认值为3600。
  """
  def assume_role(opts \\ []) do
    arguments =
      "AssumeRole"
      |> _format_arguments(opts[:app])
      |> Map.merge(%{
        "RoleArn" => opts[:app] |> _get_arn(opts[:role_arn]),
        "RoleSessionName" => opts[:role_session_name],
        "policy" => opts[:policy] |> url_encode(),
        "DurationSeconds" => opts[:duration_seconds] || 3600
      })

    signature = arguments |> _get_signature(opts[:app], "GET")
    arguments = arguments |> Map.put("Signature", signature)

    ("https://" <> AliyunAPI.Env.sts_endpoint(opts[:app]) <> "/?#{URI.encode_query(arguments)}")
    |> AliyunAPI.http_request(:get)
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, body: body, headers: headers}} ->
        response_data = Jason.decode!(body, keys: :atoms)

        {:ok, 200,
         %{
           data: %{
             request_id: response_data[:RequestId],
             assumed_role_user: response_data[:AssumedRoleUser],
             credentials: response_data[:Credentials]
           },
           body: body,
           headers: headers
         }}

      {:ok, %HTTPoison.Response{status_code: status_code, body: body, headers: headers}} ->
        response_data = Jason.decode!(body, keys: :atoms)

        {:error, status_code,
         %{
           data: %{
             message: response_data[:Message],
             code: response_data[:Code],
             request_id: response_data[:RequestId],
             host_id: response_data[:HostId]
           },
           body: body,
           headers: headers
         }}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  @doc """
  通过该接口，可以获取当前调用者的身份信息。

  https://help.aliyun.com/document_detail/28763.html

  ## opts

  """
  def get_caller_identity(opts \\ []) do
    arguments =
      "GetCallerIdentity"
      |> _format_arguments(opts[:app])
      |> Map.merge(%{})

    signature = arguments |> _get_signature(opts[:app], "GET")
    arguments = arguments |> Map.put("Signature", signature)

    ("https://" <> AliyunAPI.Env.sts_endpoint(opts[:app]) <> "/?#{URI.encode_query(arguments)}")
    |> AliyunAPI.http_request(:get)
    |> case do
      {:ok, %HTTPoison.Response{status_code: 200, body: body, headers: headers}} ->
        response_data = Jason.decode!(body, keys: :atoms)

        {:ok, 200,
         %{
           data: %{
             request_id: response_data[:RequestId],
             account_id: response_data[:AccountId],
             user_id: response_data[:UserId],
             arn: response_data[:Arn]
           },
           body: body,
           headers: headers
         }}

      {:ok, %HTTPoison.Response{status_code: status_code, body: body, headers: headers}} ->
        response_data = Jason.decode!(body, keys: :atoms)

        {:error, status_code,
         %{
           data: %{
             message: response_data[:Message],
             code: response_data[:Code],
             request_id: response_data[:RequestId],
             host_id: response_data[:HostId]
           },
           body: body,
           headers: headers
         }}

      {:error, %HTTPoison.Error{reason: reason}} ->
        {:error, reason}
    end
  end

  defp _get_signature(arguments, app, method) do
    args =
      arguments
      |> Enum.sort_by(fn {key, _} -> key end)
      |> Enum.map(fn {key, val} -> {url_encode(key), url_encode(val)} end)
      |> Enum.map(fn {key, val} -> "#{key}=#{val}" end)
      |> Enum.join("&")

    string_to_sign =
      "#{String.upcase(method)}" <> "&" <> url_encode("/") <> "&" <> url_encode(args)

    :sha
    |> :crypto.hmac("#{AliyunAPI.Env.sts_access_key_secret(app)}&", string_to_sign)
    |> Base.encode64()
  end

  defp _format_arguments(action, app) do
    %{
      "Format" => "JSON",
      "Version" => AliyunAPI.Env.sts_version(app),
      "SignatureMethod" => "HMAC-SHA1",
      "SignatureNonce" => AliyunAPI.SecureRandom.uuid(),
      "SignatureVersion" => AliyunAPI.Env.sts_signature_version(app),
      "Action" => action,
      "AccessKeyId" => AliyunAPI.Env.sts_access_key_id(app),
      "Timestamp" => Timex.format!(Timex.now(), "%Y-%m-%dT%H:%M:%SZ", :strftime)
    }
  end

  defp _get_arn(app, arn_name) when is_atom(arn_name) do
    AliyunAPI.Env.sts_arns(app) |> Map.get(arn_name)
  end

  defp _get_arn(_app, arn_name), do: arn_name
end
